import torch
from typing import List


def permutation_to_front(n: int, entries: List[int]):
    """Build permutation that maps given ordered entries to the front.

    This function creates a permutation on n elements that leaves each
    element alone except for those in `entries`. Those entries are moved
    to the front (i.e. the left) in the order of the list `entries`.

    Examples:
        >>> permutation_to_front(5, [2])
        [2, 0, 1, 3, 4]
        >>> permutation_to_front(5, [2, 1])
        [2, 1, 0, 3, 4]

    Permutations in PyTorch reminder:
        While permutations are naturally specified with dicts, PyTorch and
        NumPy use will things ike [3, 0, 1, 2] to indicate that object 3
        should be moved to slot 0, object 0 should be moved to slot 1, etc.
    """
    entries = entries.copy()
    arrangement = list(range(n))
    for q in entries:
        del arrangement[arrangement.index(q)]
    return entries + arrangement


def inverse_list_permutation(perm: List[int]):
    """Get the inverse permutation for a permutation formatted as a list.

    Examples:
        >>> inverse_list_permutation([1, 2, 3, 4, 0])
        [4, 0, 1, 2, 3]

    Permutations in PyTorch reminder:
        While permutations are naturally specified with dicts, PyTorch and
        NumPy use will things ike [3, 0, 1, 2] to indicate that object 3
        should be moved to slot 0, object 0 should be moved to slot 1, etc.
    """
    n = len(perm)
    inverse = [None for _ in range(n)]

    inverse_dict = {b: a for a, b in enumerate(perm)}
    for k, v in inverse_dict.items():
        inverse[k] = v

    return inverse


def int_to_bin(
        x: int,
        num_bits: int,
        str_output: bool = False,
        device: torch.device = torch.device("cpu")
):
    if x >= 2 ** num_bits:
        raise ValueError(f"Insufficient bits to store integer {x}.")
    if x < 0:
        raise ValueError(f"Expected a nonnegative integer, found {x}.")

    bin_format = [int(d) for d in str(bin(x))[2:]]
    padding_size = num_bits - len(bin_format)
    bin_format = padding_size * [0] + bin_format

    output = torch.tensor(bin_format, dtype=torch.bool, device=device)
    if not str_output:
        return output
    else:
        return bool_to_strings(output)


def bool_to_strings(bool_data):
    """Convert a vector of [True, False, False, True, ...] to '1001...' ."""
    def conversion(entry):
        if isinstance(bool_data, torch.Tensor):
            entry = entry.item()
        return str(int(entry))
    mapping = map(conversion, bool_data)

    return ''.join(mapping)



from .shapes import StateLayout, StateShapeError
from .shapes import count_qubits
from .shapes import count_qubits_tensor
from .shapes import count_qubits_gate_matrix
from .shapes import hilbert_space_dim
from .shapes import count_batch_dims_tensor
from .shapes import subset_roll_to_back, subset_roll_to_front
from .conversions import to_tensor_layout, to_vector_layout
from .innerprod import inner_product, norm_squared
from .innerprod import abs_squared
from .innerprod import diag_expectation_value



import torch


from unitair.simulation.operations import apply_phase
from unitair.simulation.operations import apply_operator
from unitair.simulation.operations import act_first_qubits
from unitair.simulation.operations import apply_all_qubits
from unitair.simulation.operations import swap, permute_qubits

from unitair.states import count_qubits, count_qubits_gate_matrix
from math import pi

from hypothesis import given, assume
import hypothesis.strategies as st

from tests.hypothesis_strategies.specialized import state_and_phase_angles
from tests.hypothesis_strategies.specialized import operator_and_state

from tests.hypothesis_strategies.tensors import operators, state_vectors
from unitair.simulation.utils import gate_batch_size


@given(state_and_phase_angles())
def test_apply_phase(data):
    angles = data['angles']
    original_state = data['state_vector']

    state = apply_phase(angles, original_state)
    state = apply_phase(-angles, state)
    assert state.isclose(original_state, atol=1e-4).all()

    complement_angles = 2 * pi - angles
    state = apply_phase(angles, original_state)
    state = apply_phase(complement_angles, state)
    # The weird atol here stops an issue with large angles ruining accuracy.
    # I think this is occurring because pi does not have sufficient precision.
    assert state.isclose(
        original_state,
        atol=1e-4 + angles.abs().sum() * .001
    ).all()


@given(
    op_and_state=operator_and_state(max_abs=10.),
)
def test_apply_operator_matches_act_first_qubits(op_and_state):
    op_data = op_and_state['operator_data']
    state_data = op_and_state['state_data']

    operator = op_data['operator']
    state_vector = state_data['state_vector']
    op_num_qubits = op_data['num_qubits']

    result_act_first_qubits = act_first_qubits(
        operator=operator,
        state=state_vector,
    )
    result_apply_operator = apply_operator(
        operator=operator,
        qubits=range(op_num_qubits),
        state=state_vector,
    )

    assert torch.isclose(
        result_act_first_qubits,
        result_apply_operator
    ).all()


@given(
    op_and_state=operator_and_state(),
    indices=st.lists(st.integers(min_value=0))
)
def test_act_first_qubits_batching(op_and_state, indices):
    indices = torch.Size(indices)
    op_data = op_and_state['operator_data']
    state_data = op_and_state['state_data']

    operator = op_data['operator']
    state_vector = state_data['state_vector']
    op_batch_dims = op_data['batch_dims']
    state_batch_dims = state_data['batch_dims']

    out = act_first_qubits(
        operator=operator,
        state=state_vector,
    )

    # If there are no state batch dims, there's nothing to check
    # although the above is still a smoke-test.
    if len(state_batch_dims) == 0 and len(op_batch_dims) == 0:
        return

    # When both state and op batch dims are nontrivial, they are already
    # identical because of the `operator_and_state` strategy's behavior
    if len(state_batch_dims) > 0:
        batch_dims = state_batch_dims
    elif len(op_batch_dims) > 0:
        batch_dims = op_batch_dims
    else:
        assert False, 'This code should be unreachable.'

    assume(len(indices) == len(batch_dims))

    assume(all(i < j for i, j in zip(indices, batch_dims)))
    out_entry = out[indices]

    if len(state_batch_dims) > 0:
        state_vector_entry = state_vector[indices]
    else:
        state_vector_entry = state_vector

    if len(op_batch_dims) > 0:
        operator_entry = operator[indices]
    else:
        operator_entry = operator

    unbatched_out = act_first_qubits(
        operator=operator_entry,
        state=state_vector_entry,
    )
    assert torch.isclose(out_entry, unbatched_out).all()


@given(
    op_and_state=operator_and_state(op_max_num_qubits=1, max_abs=10.),
    indices=st.lists(st.integers(min_value=0))
)
def test_apply_all_qubits_batching(op_and_state, indices):
    indices = torch.Size(indices)
    op_data = op_and_state['operator_data']
    state_data = op_and_state['state_data']

    operator = op_data['operator']
    state_vector = state_data['state_vector']
    op_batch_dims = op_data['batch_dims']
    state_batch_dims = state_data['batch_dims']

    out = apply_all_qubits(
        operator=operator,
        state=state_vector,
    )

    # If there are no state batch dims, there's nothing to check
    # although the above is still a smoke-test.
    if len(state_batch_dims) == 0:
        return

    assume(len(indices) == len(state_batch_dims))
    assume(all(i < j for i, j in zip(indices, state_batch_dims)))
    out_entry = out[indices]

    state_vector_entry = state_vector[indices]

    if len(op_batch_dims) > 0:
        operator_entry = operator[indices]
    else:
        operator_entry = operator

    unbatched_out = apply_all_qubits(
        operator=operator_entry,
        state=state_vector_entry,
    )
    assert torch.isclose(out_entry, unbatched_out).all()


@given(
    state=state_vectors(max_num_qubits=8),
    i=st.integers(min_value=0, max_value=8),
    j=st.integers(min_value=0, max_value=8)
)
def test_swap(state, i, j):
    n = count_qubits(state)

    # These ensure n >= 2 without further assumption.
    assume(i != j)
    assume(i < n)
    assume(j < n)

    state_swapped = swap(state, qubit_pair=(i, j))
    state_swapped_other_way = swap(state, qubit_pair=(j, i))

    state_double_swapped = swap(state_swapped, qubit_pair=(i, j))

    # test that swap is involutory
    assert (state == state_double_swapped).all()

    # test that swap is index-symmetrical
    assert (state_swapped == state_swapped_other_way).all()

    # check that swap matches permute
    perm = list(range(n))
    perm[i] = j
    perm[j] = i
    state_permuted = permute_qubits(permutation=perm, state_vector=state)
    assert torch.isclose(
        state_swapped,
        state_permuted,
        # atol=1e-5,
        # rtol=1e-4
    ).all()









from random import shuffle
from typing import List

import torch
from hypothesis import given, assume
import hypothesis.strategies as st
from hypothesis.strategies import composite

from unitair import utils
MAX_BITS = 10


@composite
def list_permutations(draw, min_length=1, max_length=100):
    """Draws list-format permutations"""
    length = draw(st.integers(min_value=min_length, max_value=max_length))

    perm = draw(st.lists(
        min_size=length,
        max_size=length,
        elements=st.integers(min_value=0, max_value=length - 1),
        unique=True
    ))
    return perm


@given(permutation=list_permutations())
def test_inverse_permutation_inverts(
        permutation: List[int]
):
    def permute(p: List[int], inp: list):
        """Apply list permutation p to inp."""
        return [inp[p[i]] for i in range(len(p))]

    n = len(permutation)
    inv_permutation = utils.inverse_list_permutation(permutation)
    assert len(inv_permutation) == n

    test_objects = list(range(n))
    shuffle(test_objects)

    objects_permuted = permute(permutation, test_objects)
    permute_then_inverse = permute(inv_permutation, objects_permuted)

    assert permute_then_inverse == test_objects


@given(x=st.integers(0, 2 ** (MAX_BITS - 1)), num_bits=st.integers(1, MAX_BITS))
def test_int_to_bin(x: int, num_bits: int):
    assume(x < 2 ** num_bits)
    binary_digits = utils.int_to_bin(x, num_bits=num_bits)

    powers = torch.tensor([2**n for n in reversed((range(num_bits)))])
    assert x == (binary_digits * powers).sum()






from collections import OrderedDict
from typing import Optional

import torch
from unitair.states import innerprod, count_qubits
from unitair.utils import int_to_bin


def measure(
        state: torch.Tensor,
        num_samples: int,
        raw_output: bool = False
):
    """Draw samples from the probability distribution of a given state.

    This is a very basic measurement simulation. The argument `psi` is a
    quantum state in vector layout from which the probability distribution
    for measuring in the computational basis is determined. We then draw
    the specified number of samples.

    This measurement function does not simulate the effect of measurement
    on the quantum state itself. Such simulation is not currently planned
    for unitair.

    By default, this function returns a MeasurementHistogram object
    which stores a sorted histogram of all measurements. However,
    see the raw_output argument.

    Args:
        state: The state to be measured in vector layout.
        num_samples: The number of samples to take (the state is unchanged
            after each measurement).

        raw_output: When False, the output is a more "human-readable" format.
            When True, output is a simple dict {i: count(i)} where i is an
            integer rather than a binary string. The integer i is the number
            corresponding to the binary string. For example, instead of
            '0011', we would have the int i = 3.
    """

    num_qubits = count_qubits(state)
    probs = innerprod.abs_squared(state)
    dist = torch.distributions.Categorical(probs=probs)
    samples = dist.sample([num_samples])

    unsorted_histogram = dict()
    distinct_configurations = set(samples.tolist())

    for config in distinct_configurations:
        count = (samples == config).sum().item()
        if raw_output:
            config_specification = config
        else:
            config_specification = int_to_bin(
                config, num_bits=num_qubits, str_output=True
            )
        unsorted_histogram[config_specification] = count

    if raw_output:
        return unsorted_histogram

    configs, counts = zip(*unsorted_histogram.items())
    sorted_counts, permutation = torch.sort(torch.tensor(counts),
                                            descending=True)

    histogram = MeasurementHistogram(num_qubits=num_qubits)
    for index in permutation:
        histogram[configs[index]] = counts[index]

    return histogram


class MeasurementHistogram:
    """Record of a measurement of a given number of qubits."""
    def __init__(
            self,
            num_qubits: int,
            histogram: Optional[OrderedDict] = None
    ):
        """

        """
        if histogram is None:
            histogram = OrderedDict()

        self.num_qubits = num_qubits
        self.histogram = histogram
        self._int_key_histogram = None

    def __getitem__(self, item: str):
        try:
            return self.histogram.__getitem__(item)
        except KeyError:
            if set(item).issubset({'0', '1'}) and len(item) == self.num_qubits:
                return 0
            else:
                raise KeyError(
                    f'Given key {item} is not a valid binary string for '
                    f'{self.num_qubits} bits.'
                )

    def __setitem__(self, key, value):
        self.histogram.__setitem__(key, value)

    @property
    def num_distinct_samples(self):
        return len(self.histogram)

    @property
    def observed_samples(self):
        return set(self.histogram)

    def int_key_histogram(self):
        if self._int_key_histogram is None:
            self._int_key_histogram = OrderedDict()
            factors = [
                2 ** (self.num_qubits - i - 1) for i in range(self.num_qubits)
            ]
            for config, count in self.histogram.items():
                int_key = sum([int(x) * f for x, f in zip(config, factors)])
                self._int_key_histogram[int_key] = count
        return self._int_key_histogram

    def __repr__(self):
        if self.num_distinct_samples < 50:
            short_hist = list(self.histogram.items())
            shortened = False
        else:
            short_hist = list(self.histogram.items())[:15]
            shortened = True
        short_hist = OrderedDict(short_hist)

        out = ''
        for config, count in short_hist.items():
            out += config + ': ' + str(count) + '\n'
        if shortened:
            out += f' ... ({self.num_distinct_samples - 15} lines omitted)'
        else:
            out = out[:-1]  # clip an extra \n
        return out



import warnings
from typing import Iterable, Tuple, Union, Optional
import torch
import unitair.states as states
from .utils import count_gate_batch_dims
from unitair.utils import permutation_to_front, inverse_list_permutation
import numpy as np
from unitair.gates.matrix_algebra import fuse_single_qubit_operators
from ..gates import hadamard

# TODO: should most validation be in the "front-end" vector functions or should
#  it be in the tensor functions? Many cases here actually validate in both.


def apply_phase(angles: torch.Tensor, state: torch.Tensor):
    """Multiply the kth component of state by e^(-i angles_k).

    Batching for this function follows PyTorch multiplication batching.
    Here are some examples:

    If `angles` is a scalar (a Tensor with only one element) and `state`
    is a state in vector layout with arbitrary batch dimensions, then the
    angle is applied to every component of every state.

    If `angles` is a vector with size (2^n) (where n is the number of qubits)
    and `state` is a Tensor with arbitrary batch dimensions, then
    the specified phase angle for each component is applied to each state in
    the batch of states.

    When `angles` has the same batch dimensions as `state`, we apply phase
    angles to states in one-to-one correspondence.

    To apply a batch of phases to a batch of states, but to use the same
    angle for each component of a given state, we can give `angles` size
    (*batch_dims, 1) while `state` will have size (*batch_dims, 2^n).

    Args:
        angles: Size (*angle_batch_dims, Hilbert space dim)
        state: Size (*state_batch_dims, Hilbert space dim)
    """
    phase_factors = torch.exp(-1.j * angles)
    return phase_factors * state


def apply_operator(
        operator: torch.Tensor,
        qubits: Iterable[int],
        state: torch.Tensor,
):
    """Apply an operator to a state in vector layout.

    This is a versatile function for applying arbitrary matrices to
    arbitrary subsets of qubits with optional batching behavior.

    Consider an example of acting with a two-qubit gate
    on the last two qubits of a state for three qubits.
    Let `psi` be state for three qubits with no batch dimensions so that
    `psi` has size (2, 8). (The first dimension with length 2 is for the real
    and complex parts of the state.)

    The operator that wish to apply is a 4 by 4 complex matrix (4 because
    4 = 2^(number of qubits that the gate acts on)). With Unitair
    conventions, this is constructed by a tensor with size (2, 4, 4).
    The first dimension is for the real and imaginary parts, while the
    last two are for the matrix. For example, the CNOT gate is

    operator = torch.tensor(
                 [[[1., 0., 0., 0.],
                   [0., 1., 0., 0.],
                   [0., 0., 0., 1.],
                   [0., 0., 1., 0.]],

                  [[0., 0., 0., 0.],
                   [0., 0., 0., 0.],
                   [0., 0., 0., 0.],
                   [0., 0., 0., 0.]]])

    To apply our operator to the desired qubits, we use

        apply_operator(operator, qubits=(1, 2), state=psi)

    which makes `operator` act on qubits 1 and 2 (in that order!) in
    the state `psi`. Note that qubit counting starts at 0, so the operator
    acts on the middle and final qubit. We could have used qubits=(0, 1)
    to act on the first pair, and we could have used qubits=(1, 0) to act
    on the first pair but in the reversed order.

    There are three batch structures allowed for the inputs with different
    resulting behaviors. These are very important to understand to get
    the benefit of vectorization and CUDA, it is is available.

    1. `operator` and `state` have identical batch dimension. In this case
        each operator will act on each corresponding state. The output
        will have the size (*batch, 2, 2^n) when the field is complex.
        Each batch entry of the output is the action of the corresponding
        batch entry of `operator` on the batch entry of `state`.
        Note that completely unbatched situation is a special case of this.

    2. `operator` has no batch dimension but `state` does. In this case
        there is one operator that acts on a batch of states. The output
        will have size (*batch, 2, 2^n) when the field is complex. Each
        batch entry of the output is the action of the one operator
        on the corresponding batch entry of `state`.

    3. `operator` has dimensions but `state` does not. In this case
        there are many operators that act on a single state. This function
        computes the actions of all of these operators on the one state
        in parallel--not in series. We don't compose operators
        and apply one after the other as in a quantum circuit.
        The output will have size (*batch, 2, 2^n) when the field is complex.
        Each batch entry of the output is the action of the corresponding
        operator on `state`, the one initial state.

    Args:
        operator: "Matrix" with size (*operator_batch, 2^k, 2^k). Here, k is
        the number of qubits that the operator acts on.

        qubits: Sequence of qubits for the operator to act on. This should
            have length k (same k as above). Note that the order of items in
            `qubits` is important.

        state: State in vector layout with size (*state_batch, 2^n, 2^n)
            in the complex case and (*state_batch, 2^n, 2^n) in the real case.
            Here, n is the number of qubits.
    """
    num_qubits = states.count_qubits(state)
    qubits = list(qubits)
    if not set(qubits).issubset(range(num_qubits)):
        raise ValueError(
            f'qubits={qubits} is not consistent with state vector with '
            f'{num_qubits} qubits.'
        )
    op_num_qubits = states.count_qubits_gate_matrix(operator)
    if len(qubits) != op_num_qubits:
        raise ValueError(
            f'Cannot apply operator with {op_num_qubits} to the {len(qubits)} '
            f'qubit sequence {qubits}.'
        )

    state_tensor = states.to_tensor_layout(state)
    state_tensor = apply_operator_tensor(
        operator=operator,
        qubits=qubits,
        state_tensor=state_tensor,
        num_qubits=num_qubits,
        operator_num_qubits=op_num_qubits
    )
    return states.to_vector_layout(state_tensor, num_qubits=num_qubits)


def apply_operator_tensor(
        operator: torch.Tensor,
        qubits: Iterable[int],
        state_tensor: torch.Tensor,
        num_qubits: int,
        operator_num_qubits: Optional[int] = None
):
    qubits = list(qubits)
    if operator_num_qubits is None:
        operator_num_qubits = states.count_qubits_gate_matrix(operator)
    perm = permutation_to_front(num_qubits, qubits)
    inv_perm = inverse_list_permutation(perm)

    # Permute qubits to the front
    state_tensor = permute_qubits_tensor(
        permutation=perm,
        state_tensor=state_tensor,
        num_qubits=num_qubits,
        contiguous_output=True
    )

    # Apply operator on the front consecutive qubits
    state_tensor = act_first_qubits_tensor(
        operator=operator,
        state_tensor=state_tensor,
        num_qubits=num_qubits,
        gate_num_qubits=operator_num_qubits
    )

    # Invert the earlier permutation
    return permute_qubits_tensor(
        permutation=inv_perm,
        state_tensor=state_tensor,
        num_qubits=num_qubits,
        contiguous_output=True
    )


def act_last_qubit(
        single_qubit_operator: torch.Tensor,
        state: torch.Tensor,
) -> torch.Tensor:
    """Apply an operator to the last qubit.

    Args:
        single_qubit_operator: Tensor with size (2, 2, 2) or (2, 2) in
            the complex and real cases respectively. For the complex case,
            the first dimension is for real and imaginary parts.

        state: State or batch of states in vector layout.
    """
    warnings.warn(
        'act_last_qubit and act_last_qubit_tensor are outdated. These\n'
        'functions will be removed from Unitair in a later release or may\n'
        'be revised to meet our current standards.\n'
        'Please consider using apply_operator instead.'
    )
    num_qubits = states.count_qubits(state)
    state_tensor = states.to_tensor_layout(state)
    state_tensor = act_last_qubit_tensor(
        single_qubit_operator, state_tensor
    )
    return states.to_vector_layout(state_tensor, num_qubits)


def act_last_qubit_tensor(
        single_qubit_operator: torch.Tensor,
        state_tensor: torch.Tensor,
) -> torch.Tensor:
    """Apply an operator to the last qubit of a state in tensor layout.

        Args:
            single_qubit_operator: Tensor with size (2, 2, 2) or (2, 2) in
                the complex and real cases respectively. For the complex case,
                the first dimension is for real and imaginary parts.

            state_tensor: State or batch of states in tensor layout.
        """
    def act(matrix, tensor):
        """Contract matrix with the last index of tensor."""
        return torch.einsum('ab, ...b -> ...a', matrix, tensor)

    return act(single_qubit_operator, state_tensor)


def act_first_qubits(
        operator: torch.Tensor,
        state: torch.Tensor,
):
    """Apply a multi-qubit gate to the first qubits of a state."""
    num_qubits = states.count_qubits(state)
    gate_num_qubits = states.count_qubits_gate_matrix(operator)
    if num_qubits < gate_num_qubits:
        raise ValueError(
            f'Attempted to apply a {gate_num_qubits}-qubit gate to '
            f'{num_qubits} qubit(s).'
        )
    state_tensor = states.to_tensor_layout(state)
    state_tensor = act_first_qubits_tensor(
        operator=operator,
        state_tensor=state_tensor,
        num_qubits=num_qubits,
        gate_num_qubits=gate_num_qubits
    )
    return states.to_vector_layout(state_tensor, num_qubits=num_qubits)


def act_first_qubits_tensor(
        operator: torch.Tensor,
        state_tensor: torch.Tensor,
        num_qubits: int,
        gate_num_qubits: Optional[int] = None,
):
    """Apply operator on first consecutive qubits of a state in tensor layout.

    `operator` represents an operator or batch of operators that act on
    k qubits (with k <= the number of qubits for the state).

    When used without batches, `operator` is a single-qubit operator specified
    by a tensor of size (2^k, 2^k). `state_tensor` is a state
    in tensor layout for n qubits. The operator acts on the first k consecutive
    qubits. A new state in tensor layout is then returned.

    Both operator and state_tensor can have batch dimensions, but batch
    dimensions must be compatible.

    Valid batch structures:
       `operator` and `state_tensor` have the same batch dimensions:
           In this case, each batch entry of `operator` acts on the
           corresponding entry of `state_tensor`.

       `operator` has no batch dimensions but `state_tensor` does:
           In this case, the same operator acts on every state_tensor
           in the batch.

        `operator` has batch dimensions but `state_tensor` does not:
            In this case, each operator acts on the same state.
    """
    if gate_num_qubits is None:
        gate_num_qubits = states.count_qubits_gate_matrix(operator)

    gate_dim = 2 ** gate_num_qubits

    # Determine the batch structure for the state(s) and operator(s).
    state_n_batch_dims = states.count_batch_dims_tensor(
        state_tensor, num_qubits
    )
    op_n_batch_dims = count_gate_batch_dims(operator)
    state_batch_dims = state_tensor.size()[:state_n_batch_dims]
    op_batch_dims = operator.size()[:op_n_batch_dims]

    # In the case of batched operators and a single state, we expand the state
    # into a repeated batch so that all operators act on the same state.
    if op_n_batch_dims > 0 and state_n_batch_dims == 0:
        state_tensor = state_tensor.expand(
            *op_batch_dims, *state_tensor.size()
        )
        state_n_batch_dims = op_n_batch_dims
        state_batch_dims = op_batch_dims

    state_tensor = states.subset_roll_to_back(state_tensor, state_n_batch_dims)
    operator = states.subset_roll_to_back(operator, op_n_batch_dims)

    def act(op, tensor):
        old_size = tensor.size()
        new_size = (
                (gate_dim,)
                + (num_qubits - gate_num_qubits) * (2,)
                + state_batch_dims
        )
        tensor_view = tensor.view(new_size)
        result = torch.einsum('ab..., b... -> a...', op, tensor_view)
        return result.view(old_size)

    result_batch_flipped = act(operator, state_tensor)
    return states.subset_roll_to_front(
        tensor=result_batch_flipped,
        subset_num_dims=state_n_batch_dims
    )


def apply_all_qubits(
        operator: torch.Tensor,
        state: torch.Tensor,
) -> torch.Tensor:
    """Apply the same single-qubit operator to each qubit of specified state.

    Batching cases:
        `operator` and `state` have the same batch dimensions:
            In this case, each batch entry of `operator` acts on the
            first qubit of the corresponding batch entry of `state`.

        `operator` has no batch dimensions but `state` does:
            In this case, the same operator acts on every state
            in the batch. In fact, this means that the same operator acts
            on every qubit of every entry in `state`

    Args:
        operator: Tensor with size (*batch_dims, 2, 2) defining a 2 by 2 matrix
            which will act on every qubit.

        state: State in vector layout. This means that the state is a
            tensor with size (*batch_dims, 2^num_bits,).
    """
    if states.count_qubits_gate_matrix(operator) != 1:
        raise ValueError(
            f'Expected operator on 1 qubit, found a '
            f'{states.count_qubits_gate_matrix(operator)} qubit operator.'
        )

    num_qubits = states.count_qubits(state)
    state_tensor = states.to_tensor_layout(state)
    state_tensor = apply_all_qubits_tensor(
        operator, state_tensor, num_qubits=num_qubits
    )
    return states.to_vector_layout(state_tensor, num_qubits=num_qubits)


def apply_all_qubits_tensor(
        operator: torch.Tensor,
        state_tensor: torch.Tensor,
        num_qubits: int,
):
    """Apply one single-qubit operator to each qubit of state in tensor layout.

    Batching cases:
        `operator` and `state_tensor` have the same batch dimensions:
            In this case, each batch entry of `operator` acts on the
            first qubit of the corresponding batch entry of `state_tensor`.

        `operator` has no batch dimensions but `state_tensor` does:
            In this case, the same operator acts on every state
            in the batch. In fact, this means that the same operator acts
            on every qubit of every entry in `state_tensor`

    Args:
        operator: Tensor with size (*batch_dims, 2, 2) giving a 2 by 2 matrix.

        state_tensor: State in tensor layout. Size is
            (*batch_dims, 2, 2, ...,2, 2) where the number of 2's is equal to
            the number of qubits.

        num_qubits: The number of qubits for the quantum state.
    """
    state_tensor = act_first_qubits_tensor(
        operator=operator,
        state_tensor=state_tensor,
        num_qubits=num_qubits,
        gate_num_qubits=1
    )
    for i in range(1, num_qubits):
        state_tensor = swap_tensor(
            state_tensor, qubit_pair=(0, i), num_qubits=num_qubits
        )
        state_tensor = act_first_qubits_tensor(
            operator, state_tensor, num_qubits=num_qubits, gate_num_qubits=1
        )

    state_tensor = roll_qubits_tensor(
        state_tensor, num_qubits, num_steps=-1
    )

    return state_tensor


def apply_to_qubits(
        operators: Iterable[torch.Tensor],
        qubits: Iterable[int],
        state: torch.Tensor,
):
    """Apply single qubit gates to specified qubits of a state.

    This function applies an iterable of single-qubit operators to
    the qubits specified by an iterable of integers. Qubit indices
    must range from 0 to n-1 where n is the number of qubits.

    We collect all operators acting on a given qubit and multiply them prior
    to acting on the state; this "gate fusion" is often more efficient than
    applying gates to a state one-by-one.

    Args:
        operators: Iterable over single qubit operators. Size of each element
            should be (*batch_dims, 2, 2). All batch dims must be the same.

        qubits: Iterable over the qubit integers for each operator. `operators`
            and `qubits` must line up in the sense that zip(operators, qubits)
            produces matrix, qubit pairs appropriately.

        state: State in vector layout. The size should be (*batch_dims, 2^n). `
    """
    num_qubits = states.count_qubits(state)
    state_tensor = states.to_tensor_layout(state)
    state_tensor = apply_to_qubits_tensor(
        operators, qubits, state_tensor, num_qubits
    )
    return states.to_vector_layout(state_tensor, num_qubits=num_qubits)


def apply_to_qubits_tensor(
        operators: Iterable[torch.Tensor],
        qubits: Iterable[int],
        state_tensor: torch.Tensor,
        num_qubits: int,
):
    """Apply single qubit gates to specified qubits of state in tensor layout.

    This function collects all operators acting on a given qubit and multiplies
    them prior to acting on the state; this is often more efficient.

    Args:
        operators: Iterable over single qubit operators. Size of each element
            should be (*batch_dims, 2, 2). All batch dims must be the same.

        qubits: Iterable over the qubit integers for each operator. `operators`
            and `qubits` must line up in the sense that zip(operators, qubits)
            produces matrix, qubit pairs appropriately.

        state_tensor: State in tensor layout.

        num_qubits: The number of qubits for the state.    `
    """
    # First we fuse gates acting on the same qubit.
    qubits_to_fused_ops = fuse_single_qubit_operators(qubits, operators)

    # If qubit 0 is included, apply its gate first to avoid back and forth
    # permutation. Note that we pop the gate so qubit 0 will not be used again.
    try:
        gate = qubits_to_fused_ops.pop(0)
        state_tensor = act_first_qubits_tensor(
            gate,
            state_tensor,
            num_qubits,
            gate_num_qubits=1
        )
    except KeyError:
        pass

    perm = list(range(num_qubits))
    for q, gate in qubits_to_fused_ops.items():

        # Track the overall permutation as we swap qubits.
        # Very importantly, each q is encountered either once or never.
        perm[q], perm[0] = perm[0], q
        state_tensor = swap_tensor(state_tensor, (0, q), num_qubits)

        state_tensor = act_first_qubits_tensor(
            gate, state_tensor, num_qubits, gate_num_qubits=1
        )
    return permute_qubits_tensor(
        permutation=inverse_list_permutation(perm),
        state_tensor=state_tensor,
        num_qubits=num_qubits
    )


def swap(state: torch.Tensor, qubit_pair: Tuple[int, int]):
    """Swap a pair of qubits for a state.

    This operation is more natural for states in tensor layout, and in fact
    this function wraps `swap_tensor`.
    """
    num_qubits = states.count_qubits(state)
    state_tensor = states.to_tensor_layout(state)
    state_tensor = swap_tensor(
        state_tensor, qubit_pair, num_qubits
    )
    return states.to_vector_layout(state_tensor, num_qubits)


def swap_tensor(
        state_tensor: torch.Tensor,
        qubit_pair: Tuple[int, int],
        num_qubits: int,
):
    """Swap a pair of qubits for a state in tensor layout.

    For the corresponding operation of states in vector format, see
    the function `swap_qubits`.
    """
    if qubit_pair[0] == qubit_pair[1]:
        return state_tensor
    q1, q2 = states.shapes.get_qubit_indices(
        index=qubit_pair,
        state_tensor=state_tensor,
        num_qubits=num_qubits
    )
    return state_tensor.transpose(q1, q2)


def roll_qubits(
        state: torch.Tensor, num_steps=1
):
    """Perform a cyclic permutation of qubits for a state.

    This operation is more natural for states in tensor layout, and
    this function wraps `roll_qubits_tensor`.

    This function rolls qubits forward by the specified number of
    steps. For example, is psi is a quantum state with

    < 0100 | psi > = c

    where c is some complex number, then the calling
    roll_qubits(psi, num_steps=1) will return a new state psi' that will have

    < 0010 | psi' > = c.
    """
    num_qubits = states.count_qubits(state)

    state_tensor = states.to_tensor_layout(state)
    state_tensor = roll_qubits_tensor(state_tensor, num_qubits, num_steps)
    return states.to_vector_layout(state_tensor, num_qubits)


def roll_qubits_tensor(
        state_tensor: torch.Tensor,
        num_qubits: int,
        num_steps: int = 1
):
    """Perform a cyclic permutation of qubits for a state in tensor layout.

    If the initial tensor is psi, the output of this function
    is a new tensor psi_rolled which is related to psi by

    psi_rolled[a_0, a_1, ..., a_{n-1}]
        = psi[a_k, a_{k+1}, ..., a_{n-1}, a_0, ..., a_{k-1}].

    Note that this formula does not include batch dimensions which are
    allowed and arbitrary.

    Args:
        state_tensor: state in tensor layout to be permuted.

        num_qubits: Number of qubits for the state.

        num_steps: Number of indices to cycle.
    """
    num_batch_dims = states.count_batch_dims_tensor(
        state_tensor, num_qubits)
    num_steps = num_steps % num_qubits
    if num_steps == 0:
        return state_tensor

    identity = list(range(num_batch_dims, num_batch_dims + num_qubits))
    perm = identity[-num_steps:] + identity[:-num_steps]
    perm = list(range(num_batch_dims)) + perm
    return state_tensor.permute(perm)


def permute_qubits(
        permutation: Iterable[int],
        state_vector: torch.Tensor
):
    """Permute qubits for a state in vector layout.

    If `state_vector` has batch dimensions, then the same permutation is
    applied to all batch entries. In particular, this function does not
    permute batch indices (use torch.permute for that).

    Args:
        permutation: Sequence of integers defining a permutation among qubits.
        state_vector: State in vector layout.
    """
    num_qubits = states.count_qubits(state_vector)
    state_tensor = states.to_tensor_layout(state_vector)
    permuted_state_tensor = permute_qubits_tensor(
        permutation=permutation,
        state_tensor=state_tensor,
        num_qubits=num_qubits,
    )
    return states.to_vector_layout(
        permuted_state_tensor, num_qubits=num_qubits
    )


def permute_qubits_tensor(
        permutation: Iterable[int],
        state_tensor: torch.Tensor,
        num_qubits: int,
        contiguous_output: bool = False
):
    """Permute qubits for a state in tensor layout.

    If `state_tensor` has batch dimensions, then the same permutation is
    applied to all batch entries. In particular, this function does not
    permute batch indices (use torch.permute for that).

    Args:
        permutation: Sequence of integers defining a permutation among qubits.
        state_tensor: State in vector layout.
        num_qubits: Number of qubits for the state.
        contiguous_output: When True, returns a contiguous copy of the result.
    """
    # non_qubit_dims includes the complex dimension in the complex case.
    non_qubit_dims = state_tensor.dim() - num_qubits
    identity_perm = tuple(i for i in range(non_qubit_dims))

    permutation = tuple(i + non_qubit_dims for i in permutation)

    state_tensor = torch.permute(state_tensor, identity_perm + permutation)
    if not contiguous_output:
        return state_tensor
    else:
        return state_tensor.contiguous()


def multi_cz(
        qubit_pairs: torch.Tensor,
        state_vector: torch.Tensor,
        num_bits_memory_cutoff: Optional[int] = None
):
    """Apply CZ gates to specified qubit pairs.

    This function can either apply a single CZ gate or multiple gates
    to arbitrary pairs of qubits. For example, if the parameter
    `qubit_pairs` is set to torch.tensor([0, 1]), then qubits 0 and 1
    are used as the (interchangeable) control and target and the gate
    CZ(control=0, target=1) is applied to `state_vector`. However,
    if instead we have

        qubit_pairs=torch.tensor([[0, 1], [0, 2]])

    then the gates CZ(control=0, target=1) and CZ(control=0, target=2) are
    applied to `state_vector`. (Note that the two operators commute.)

    CUDA usage note:
        This function can benefit substantially from GPU acceleration.
        This is invoked when the arguments `qubit_pairs` and `state_vector`
        are CUDA tensors. See argument `num_bits_memory_cutoff` for a
        parameter that can be tuned to optimize GPU performance.

    Args:
        qubit_pairs: Tensor with size (2,) or (num_pairs, 2) giving a
            collection of control-target pairs.

        state_vector: State in vector layout upon which to apply CZ gates.

        num_bits_memory_cutoff: This advanced parameter can be tuned to avoid
            running out of CUDA memory in the case where multiple qubit pairs
            are used. Normally multiple qubit pairs are batched and processed
            with parallelization, but if log_2(num_pairs) + num_qubits exceeds
            num_bits_memory_cutoff, we instead use a for loop. When this
            parameter is None, an estimate is used based on device properties.
    """
    num_qubits = states.count_qubits(state_vector)
    dev = qubit_pairs.device
    if qubit_pairs.dim() == 1:
        qubit_pairs = qubit_pairs.view(1, 2)

    if state_vector.device.type != qubit_pairs.device.type:
        raise ValueError(
            'state_vector and qubit_qubit_pairs must have same device.'
        )

    elif qubit_pairs.dim() != 2:
        raise ValueError('qubit_pairs must have size (2,) or (num_pairs, 2).')
    if (qubit_pairs >= num_qubits).any():
        raise ValueError(
            'Control/target indices for CZ gate must be less than num_bits.'
        )

    num_pairs = qubit_pairs.size()[0]
    if num_bits_memory_cutoff is None:
        if dev.type == 'cpu':
            num_bits_memory_cutoff = 1000  # Effectively infinity.
        else:
            mem = torch.cuda.get_device_properties(dev).total_memory
            num_bits_memory_cutoff = mem - 5

    use_batched_pairs = (
            num_qubits + np.log2(num_pairs) < num_bits_memory_cutoff
    )

    # control and target are interchangeable.
    control = qubit_pairs[:, 0]
    target = qubit_pairs[:, 1]
    if (control == target).any():
        raise ValueError('Control and target qubits are not distinct.')
    pow_c = num_qubits - control - 1
    pow_t = num_qubits - target - 1
    critical_int = 2 ** pow_c + 2 ** pow_t

    all_ints = torch.arange(2 ** num_qubits, device=dev)

    if use_batched_pairs:
        match = torch.bitwise_and(
            all_ints.view(-1, 1),
            critical_int.view(1, -1)
        ) == critical_int

        phases = torch.prod(1 - 2 * match, dim=-1)
    else:
        phases = torch.ones(2 ** num_qubits, device=dev)
        for crit in critical_int:
            match = torch.bitwise_and(all_ints, crit) == crit
            phases *= 1 - 2 * match

    return phases * state_vector


def multi_controlled_z(
        qubits: Iterable[int],
        state_vector: torch.Tensor,
):
    """Apply a CC...CZ gate to the given qubits.

    Note that controlled-Z operations are independent of the control versus
    target. The parameter qubits includes all qubits that the gate acts on.
    """
    total_num_qubits = states.count_qubits(state_vector)
    qubits = list(qubits)
    qubit_set = set(qubits)

    inert_qubits = [
        q for q in range(total_num_qubits)
        if q not in qubit_set
    ]

    # Prepare diag(1, 1, ..., 1, -1) for the subsystem only.
    subsystem_factors = torch.ones(2 ** len(qubits))
    subsystem_factors[-1] = -1.
    factors = subsystem_factors.repeat(2 ** len(inert_qubits))

    # Put unused qubits on the left and C...CZ qubits on the right.
    perm = inert_qubits + qubits
    reverse_perm = [0] * total_num_qubits
    for q, i in enumerate(perm):
        reverse_perm[i] = q

    # Permute, apply the phase,
    state_vector = permute_qubits(perm, state_vector)
    state_vector = factors * state_vector
    return permute_qubits(reverse_perm, state_vector)


def multi_controlled_x(
        state_vector: torch.Tensor,
        controls: Iterable[int],
        target: int
):
    controls = list(controls)
    h = hadamard()
    state_vector = apply_operator(
        operator=h,
        qubits=[target],
        state=state_vector
    )
    state_vector = multi_controlled_z(
        qubits=controls + [target],
        state_vector=state_vector
    )
    return apply_operator(
        operator=h,
        qubits=[target],
        state=state_vector
    )



from typing import Optional

import hypothesis.strategies as st
import numpy as np
from hypothesis import assume

from unitair import count_qubits
from unitair.simulation.utils import gate_batch_size
from .tensors import state_vectors, tensors_size_fixed, \
    state_vectors_with_metadata, operators


@st.composite
def state_and_phase_angles(
        draw,
        batch_rank_limit: int = 3,
        batch_size_limit: int = 3,
        min_num_qubits: int = 1,
        max_num_qubits: int = 8,
):
    """Draws batches of state vectors along with compatible phase angles.

    This is a specialized strategy for unitair.simulation.apply_phase.
    """
    state = draw(state_vectors(
        batch_rank_limit=batch_rank_limit,
        batch_size_limit=batch_size_limit,
        min_num_qubits=min_num_qubits,
        max_num_qubits=max_num_qubits,
        strictly_complex=True
    ))
    batch_dims = state.size()[:-1]
    num_qubits = count_qubits(state)
    shape = batch_dims + (2 ** num_qubits,)
    angles = draw(tensors_size_fixed(shape=shape, dtype=np.dtype('float32')))
    return {
        'angles': angles,
        'state_vector': state
    }


@st.composite
def operator_and_state(
        draw,
        min_num_qubits: int = 1,
        max_num_qubits: int = 8,
        op_min_num_qubits: int = 1,
        op_max_num_qubits: int = 5,
        batch_max_num_indices: int = 3,
        batch_max_index_range: int = 5,
        strictly_complex: bool = True,
        nonzero: bool = False,
        max_abs: Optional[float] = None
):
    """Draws batches of operators and states that are compatible.

    By "compatible", we mean that either the operator has no batch
    dimensions or that the operator and state have identical batch
    dimensions.
    """
    state_data = draw(state_vectors_with_metadata(
        batch_rank_limit=batch_max_num_indices,
        batch_size_limit=batch_max_index_range,
        min_num_qubits=min_num_qubits,
        max_num_qubits=max_num_qubits,
        strictly_complex=strictly_complex
    ))

    state_num_qubits = state_data['num_qubits']
    state_batch_dims = state_data['batch_dims']
    state_dtype = state_data['dtype']

    assume(
        state_num_qubits >= op_min_num_qubits
    )

    op_num_qubits = draw(st.integers(
        min_value=op_min_num_qubits,
        max_value=min(state_num_qubits, op_max_num_qubits)
    ))

    operator = draw(operators(
        min_num_qubits=op_num_qubits,
        max_num_qubits=op_num_qubits,
        batch_max_num_indices=batch_max_num_indices,
        batch_max_index_range=batch_max_index_range,
        forced_dtype=state_dtype,
        nonzero=nonzero,
        max_abs=max_abs
    ))
    op_batch_dims = gate_batch_size(gate=operator)

    assume(
        len(op_batch_dims) == 0
        or len(state_batch_dims) == 0
        or op_batch_dims == state_batch_dims
    )

    op_data = dict(
        operator=operator,
        num_qubits=op_num_qubits,
        batch_dims=op_batch_dims,
        dtype=state_dtype
    )

    return dict(operator_data=op_data, state_data=state_data)





import torch


def abs_squared(state: torch.Tensor):
    """Compute the vector of measurement probabilities for state.

    For an $N$-dimensional Hilbert space, in a given basis a state
    can be regarded as a vector (x_1, ..., x_N). This function
    simply computes (|x_1|^2, ..., |x_N|^2).

    When state is a batch, the vector of absolute value squared coefficients
    is computed for each batch entry.

    Note:
        This function trivially wraps torch operations. We recommend its usage
        when it improves readability.

    Args:
        state (Tensor): State or batch of states in vector layout. Tensor size
            should be (*batch_dims, N). State can be real or complex.

    Returns:
        Tensor with size (*batch_dims, N) giving all measurement
        probabilities for all states in the batch.
    """
    return (state.conj() * state).real


def norm_squared(state: torch.Tensor):
    """Compute the L^2 norm-squared < state | state >.

    `state` can be a batch of states. The L^2 norm is used in the real and
    complex case.

    When `state` is normalized, norm_squared(state) should return 1.

    Args:
        state (Tensor): State or batch of states in vector layout. Tensor size
            can be (*batch_dims, N) where N is typically 2^(num_qubits). State
            dtype can be real or complex.

    Returns:
        Tensor with size (*batch_dims,) giving the squared norm of every state
        in the batch.
    """
    return torch.sum(abs_squared(state), dim=-1)


def diag_expectation_value(diag_values, state):
    """Get the expectation value of a diagonal matrix.

    Args:
        diag_values: Tensor with size (*batch_dims, 2^n) where n is the
            number of qubits. For a given batch entry, the 2^n elements
            are the diagonal matrix elements of the operator.

        state: State in vector layout.
    """
    return torch.sum(abs_squared(state) * diag_values, dim=-1)


def inner_product(state_1, state_2):
    """Compute < state_1 | state_2 > (the left entry is conjugate-linear).
    """
    return torch.sum(state_1.conj() * state_2, dim=-1)



from typing import Union, Sequence, Optional

import torch

from . import states


def unit_vector_from_bitstring(
        bitstring: Union[str, Sequence[int]],
        device: torch.device = torch.device("cpu"),
        dtype: torch.dtype = torch.complex64
):
    """Create a unit vector from a classical bit specification.

    This is meant for convenience in interactive experimentation.
    The function `unit_vector` is preferable in many contexts.

    Examples:
        >>> unit_vector_from_bitstring('000')  # doctest: +SKIP
        tensor([[1., 0., 0., 0., 0., 0., 0., 0.],
                [0., 0., 0., 0., 0., 0., 0., 0.]])

        Bitstring specification is somewhat flexible.
        >>> unit_vector_from_bitstring([1, 0])  # doctest: +SKIP
        tensor([[0., 0., 1., 0.],
                [0., 0., 0., 0.]])

    """
    try:
        bitstring = torch.tensor(bitstring)
    except TypeError:
        bitstring = torch.tensor([int(i) for i in bitstring])

    implied_num_bits = bitstring.size()[0]
    powers = torch.tensor([2 ** k for k in reversed(range(implied_num_bits))])
    index_implied = (powers * bitstring).sum()

    return unit_vector(
        index=index_implied,
        num_qubits=implied_num_bits,
        device=device,
        dtype=dtype
    )


def unit_vector(
        index: int,
        num_qubits: Optional[int] = None,
        dim: Optional[int] = None,
        device: torch.device = torch.device("cpu"),
        dtype: torch.dtype = torch.complex64
):
    """Create a real or complex unit vector in a Hilbert space."""
    if dim is None:
        if num_qubits is None:
            raise TypeError(
                'To specify a unit vector, provide either `num_qubits` or '
                ' `dim`.')
        dim = 2 ** num_qubits
    else:
        if num_qubits is not None:
            raise TypeError(
                'Unit vector can be specified by `dim` or `num_qubits` but '
                'not both.'
            )
    vector = torch.zeros(dim, device=device, dtype=dtype)
    vector[index] = 1
    return vector


def rand_state(
        num_qubits: int,
        batch_dims: Optional[Sequence] = None,
        device: torch.device = torch.device("cpu"),
        dtype: torch.dtype = torch.complex64,
        requires_grad: bool = False
) -> torch.Tensor:
    """Create a normalized random state in vector layout.

    States are drawn from a uniform distribution on the unit sphere embedded
    in F^N where N = 2^(num_qubits) and F the complex numbers when the
    specified dtype is complex and the reals when the dtype is real.

    When batch_dims is provided, a batch of random states is generated with
    specified shape.
    """
    if batch_dims is None:
        size = torch.Size((2 ** num_qubits,))
    else:
        size = torch.Size(batch_dims) + torch.Size((2 ** num_qubits,))

    state = torch.randn(
        size,
        device=device,
        dtype=dtype
    )
    norm = states.norm_squared(state).sqrt()
    norm_size_for_division = norm.size() + (1,)
    state /= norm.view(norm_size_for_division)
    return state.requires_grad_(requires_grad)


def uniform_superposition(
        num_qubits: int,
        batch_dims: Optional[Sequence] = None,
        device: torch.device = torch.device("cpu"),
        dtype: torch.dtype = torch.complex64,
        requires_grad: bool = False
):
    """Create the uniform superposition state |+...+>.

    This function is defined for the complex case only. Optional batch
    dimensions can be provided in which case the same uniform superposition
    is copied to form a batch with specified shape.
    """
    scaling = 2 ** (-num_qubits/2.)
    state = scaling * torch.ones(
        2 ** num_qubits,
        device=device,
        dtype=dtype
    )

    if batch_dims is not None:
        batch_dims = torch.Size(batch_dims)
        state = state.repeat(batch_dims + (1,))

    state.requires_grad_(requires_grad)
    return state



import unittest
import torch
import unitair.states as states
import random
import unitair


class VectorTensorConversions(unittest.TestCase):

    def test_vector_tensor_vector(self):
        n_min = 5
        n_max = 13

        n_qubits = random.randint(n_min, n_max)
        real_vec = torch.rand(2 ** n_qubits)
        vec_batched = torch.rand(7, 1, 4, 2 ** n_qubits)

        tens = states.to_tensor_layout(real_vec)
        tens_batched = states.to_tensor_layout(vec_batched)

        vec_back = states.to_vector_layout(
            state_tensor=tens,
            num_qubits=n_qubits
        )
        vec_back_batched = states.to_vector_layout(
            state_tensor=tens_batched,
            num_qubits=n_qubits
        )

        self.assertTrue(
            (vec_back == real_vec).all()
        )

        self.assertTrue(
            (vec_back_batched == vec_batched).all()
        )


    def test_tensor_vector_tensor(self):

        n_min = 5
        n_max = 13
        n = random.randint(n_min, n_max)
        tens = torch.rand(n * (2,))

        tens_batch = torch.rand((1, 2, 3,) + n * (2,))
        real_vec = states.to_vector_layout(tens, n)
        vec_batch = states.to_vector_layout(tens_batch, n)
        real_tens_back = states.to_tensor_layout(real_vec)
        tens_back_batch = states.to_tensor_layout(vec_batch)

        self.assertTrue((tens == real_tens_back).all())

        self.assertTrue((tens_batch == tens_back_batch).all())


class TestQubitCounting(unittest.TestCase):

    def test_count_qubits(self):
        state = torch.rand(2 ** 4)
        n = unitair.states.count_qubits(state)
        self.assertEqual(n, 4)

        state = torch.rand(7, 4, 1, 2 ** 5)
        n = unitair.states.count_qubits(state)
        self.assertEqual(n, 5)

        state = torch.rand(2, 2, 2, 2, 2)
        n = unitair.states.count_qubits(state)
        self.assertEqual(n, 1)

        state_tensor = torch.rand(4, 3, 2, 2, 2, 2, 2)
        state = unitair.states.to_vector_layout(state_tensor, num_qubits=3)
        n = unitair.states.count_qubits(state)
        self.assertEqual(n, 3)

    def test_qubit_indices(self):
        n = 4
        state = torch.rand(5, 2**n)
        ind_selected = states.shapes.get_qubit_indices(
            index=3,
            state_tensor=states.to_tensor_layout(state),
            num_qubits=n
        )
        self.assertEqual(4, ind_selected)

        n = 5
        state = torch.rand(2, 5, 7, 2 ** n)
        ind_selected = states.shapes.get_qubit_indices(
            index=torch.tensor([2, -3, 1, 0]),
            state_tensor=states.to_tensor_layout(state),
            num_qubits=n
        )
        expected = torch.tensor([5, -3, 4, 3])
        self.assertTrue(expected.eq(ind_selected).all())

        n = 4
        state = torch.rand(2, 2, 2 ** n)
        ind_selected = states.shapes.get_qubit_indices(
            index=[-2, -4, 0, 1],
            state_tensor=states.to_tensor_layout(state),
            num_qubits=n
        )
        expected = [-2, -4, 2, 3]
        self.assertEqual(expected, ind_selected)


if __name__ == '__main__':
    unittest.main()



import torch
import decorator
from typing import Optional


@decorator.decorator
def parameterized_gate(gate_function, strictly_complex: bool = False, *args):
    """Build a single qubit gate given a gate function.

    This is intended to be used as a decorator. The gate function takes
    in a tensor of parameters and returns a nested list of tensors
    or a tensor.
    # TODO: add documentation to explain this better.
    """
    params = args[0]
    squeeze = False
    if not isinstance(params, torch.Tensor):
        params = torch.tensor([float(params)], device=torch.device("cpu"))
        squeeze = True
    elif params.dim() == 0:
        params = params.unsqueeze(0)
        squeeze = True

    gate = gate_function(params)
    if not isinstance(gate, torch.Tensor):
        gate = nested_stack(gate, roll=True)
    if squeeze:
        gate = gate.squeeze(0)

    if strictly_complex:
        if not torch.is_complex(gate):
            raise TypeError(
                'This parameterized gate is required to be complex but the\n'
                f'gate ended up having dtype {gate.dtype}. Check the\n'
                f'construction of the gate decorated with parameterized_gate,\n'
                f'and ensure that it returns a complex Tensor with the given\n'
                f'parameters (which have dtype {params.dtype}).')
    return gate


@decorator.decorator
def constant_gate(gate_function, strictly_complex: bool = False, *args):
    device = args[0]
    dtype = args[1]
    if device is None:
        device = torch.device("cpu")
    gate = torch.tensor(gate_function(), device=device)
    if dtype is not None:
        gate = gate.to(dtype=dtype)
    if strictly_complex:
        if not torch.is_complex(gate):
            raise TypeError('This gate requires a complex dtype, but it ended '
                            f'up with dtype {gate.dtype}.')
    return gate


def nested_stack(params, roll: bool = False):
    """Form a tensor from a nexted list of tensors.

    This function is a generalization of torch.stack. For proper usage,
    it's important that params is a nested list with shape consistent with
    and array. The innermost elements of that nested list should be PyTorch
    tensors, all of which have identical size.

    For an example, suppose that a, b, c, and d are all tensors of size (5,).
    Then, nested_stack([[a, b], [c, d]]) returns a tensor of size (2, 2, 5).

    If roll is set to True, then the dimensions of the tensors (like a, b, c
    and d in the example above) will be permuted to the start of the output.
    This is useful if those dimensions were supposed to be batch dimensions.
    In the example, the output with roll=True would have size (5, 2, 2).
    If instead a, b, c, and d all had size (6, 9, 8), then the output size
    would be (6, 9, 8, 2, 2) if roll=True and (2, 2, 6, 9, 8) if roll=False.
    """
    def recursive_stack(params_):
        if isinstance(params_[0], torch.Tensor):
            return torch.stack(params_)
        num_rows = len(params_)
        return torch.stack(
            [nested_stack(params_[i]) for i in range(num_rows)]
        )
    stacked = recursive_stack(params).squeeze(0)

    if roll:
        inner = params[0]
        while not isinstance(inner, torch.Tensor):
            inner = inner[0]
        inner_dim = inner.dim()
        perm = list(range(stacked.dim()-inner_dim, stacked.dim())) + list(range(stacked.dim()-inner_dim))
        return stacked.permute(perm)
    else:
        return stacked



from typing import Iterable

import torch


def fuse_single_qubit_operators(
        qubits: Iterable[int],
        operators: Iterable[torch.Tensor],
):
    """Multiply together gates acting on various single qubits.

    Suppose that we have a sequence of single-qubit gates that
    should act, one after the other, on designated qubits.
    If any qubit repeats, then we can matrix multiply the operators
    to reduce to a single matrix acting on each qubit. This function
    performs this operation and collects the "fused" operations
    in a simple dict.

    If the argument `qubits` is [2, 4, 2, 7] and the `operators` is
    a list with length four as in

        A = operators[0]
        B = operators[1]
        C = operators[2]
        D = operators[3],

    then this function will return the dict
        {
            2: CA
            4: B
            7: D
        }

    where CA is the matrix multiplication of torch.matmul(C, A).

    Args:
        qubits: Iterable of integers giving the qubits that gates act on.
        operators: Iterable of Tensor objects specifying single qubit
            operators. The size should be (*batch_dims, 2, 2).

    Returns:
        Dict mapping qubits to act on to a fused gate.
    """
    qubits_to_fused_ops = dict()
    for q, op in zip(qubits, operators):
        if q in qubits_to_fused_ops:
            qubits_to_fused_ops[q] = torch.matmul(op, qubits_to_fused_ops[q])
        else:
            qubits_to_fused_ops[q] = op

    return qubits_to_fused_ops



import torch


def count_gate_batch_dims(gate: torch.Tensor):
    """Compute the number of batch dimensions for a given gate batch.

    Args:
        gate: Tensor with size (*batch_dims, 2^k, 2^k) where k is
            the number of qubits on which the gate acts.
    """
    out = gate.dim() - 2

    if out < 0:
        raise RuntimeError(
            f'Gate with size {gate.size()} is incorrectly shaped for an '
            f'operator batch. Expected size is\n'
            f'  (*optional_batch_dims, 2^k, 2^k)\n'
            f'with k the number of qubits on which the gate acts.'
        )
    return out


def gate_batch_size(gate: torch.Tensor):
    """Get the batch shape for given gate batch.

    Args:
        gate: Tensor with size (*batch_dims, 2^k, 2^k) where k is
            the number of qubits on which the gate acts.
    """
    num_batch_dims = count_gate_batch_dims(gate)
    return gate.size()[:num_batch_dims]



import torch

from unitair.initializations import rand_state, uniform_superposition
from hypothesis import given
import hypothesis.strategies as st
from tests.hypothesis_strategies import sizes, real_and_complex_torch_dtypes
from unitair.states.innerprod import norm_squared


@given(
    num_qubits=st.integers(1, 5),
    batch_dims=sizes(),
    field=st.sampled_from([torch.complex64, torch.float32])
)
def test_rand_state(num_qubits, batch_dims, field):
    state = rand_state(num_qubits=num_qubits, batch_dims=batch_dims)

    # Confirm expected batch sizes.
    expected_size = batch_dims + (2 ** num_qubits,)

    assert state.size() == expected_size

    # Check that norm_squared == 1 for all entries.
    actual = norm_squared(state)
    expected = torch.tensor(1.)
    assert (actual.isclose(expected)).all()


@given(num_qubits=st.integers(1, 5), batch_dims=sizes())
def test_uniform_superposition(num_qubits, batch_dims):
    state = uniform_superposition(
        num_qubits=num_qubits, batch_dims=batch_dims,
    )

    assert state.size() == batch_dims + (2**num_qubits,)

    expected_value_real = torch.tensor(2 ** (-num_qubits/2.))
    expected_value_imag = torch.tensor(0.)
    real, imag = state.real, state.imag
    assert real.isclose(expected_value_real).all()
    assert imag.isclose(expected_value_imag).all()



import torch
from . import shapes


def to_tensor_layout(state: torch.Tensor) -> torch.Tensor:
    """Convert a state in vector layout to one in tensor layout.

    Tensor layout is a convenient view of a state where different indices
    are used for different qubits. An n qubit state with arbitrary
    batch dimensions is represented, in tensor layout, as a torch.Tensor
    with size

        (*optional_batch_dims, 2, 2, ..., 2)

    where the number of 2's is equal to the number of qubits. (This can
    be confusing for batch dims with length 2!). Tensor layout should be
    contrasted with "vector layout" which is the standard form of states in
    Unitair. In vector layout, states have size

        (*optional_batch_dims, 2^n).

    This function converts from vector to tensor layout.

    Note:
        Even though tensor layout has the word "tensor", both tensor and
        vector layout states are torch.Tensor objects.

    Args:
        state: Tensor with size (*batch_dims, 2^num_qubits) representing a
            quantum state in vector layout.

    Returns:
        `state` viewed as a tensor with size (*batch_dims, 2, 2, ..., 2).
    """
    num_qubits = shapes.count_qubits(state)

    # We might consider adding a `reject_if_not_contiguous` optional parameter.
    # If so, this conversion to a contiguous Tensor would not happen without
    # the user being aware of it.
    if not state.is_contiguous():
        state = state.contiguous()

    qubit_size = torch.Size(num_qubits * [2])
    batch_size = (state.size())[:-1]
    return state.view(batch_size + qubit_size)


def to_vector_layout(
        state_tensor: torch.Tensor,
        num_qubits: int
) -> torch.Tensor:
    """Convert a state in tensor layout to one in vector layout.

    Vector layout is the standard form of quantum states in Unitair.
    States in vector layout are torch.Tensor objects with size

        (*optional_batch_dims, 2^n)

    where n is the number of qubits. Meanwhile, Tensor layout is a convenient
    view of a state where different indices are used for different qubits. An
    n qubit state is represented, in tensor layout, as a torch.Tensor
    with size

        (*optional_batch_dims, 2, 2, ..., 2)

    where the number of 2's is equal to the number of qubits. (This can
    be confusing for batch dims with length 2!).

    If state_tensor is not contiguous, this function will first
    get a contiguous copy. We may consider making this behavior optional
    with a `reject_if_not_contiguous` optional parameter.

    Note:
        Even though tensor layout has the word "tensor", both tensor and
        vector layout states are torch.Tensor objects.

    Args:
        state_tensor: Tensor with size (*batch_dims, 2, 2, ..., 2) where the
            number of 2's is the number of qubits.

        num_qubits: The number of qubits for the state.

    Returns:
        Tensor with size (*batch_dims, 2^num_qubits,) representing a quantum
        state in vector layout.
    """
    if not state_tensor.is_contiguous():
        state_tensor = state_tensor.contiguous()

    qubit_size = torch.Size((2 ** num_qubits,))
    batch_size = (state_tensor.size())[:-num_qubits]
    return state_tensor.view(batch_size + qubit_size)



from unitair.states.shapes import count_qubits
from tests.hypothesis_strategies import state_vectors_with_metadata
from hypothesis import given


@given(state_vectors_with_metadata())
def test_count_qubits(state_data):
    state_vector = state_data['state_vector']
    num_qubits = state_data['num_qubits']

    assert count_qubits(state_vector) == num_qubits



from typing import Optional, Tuple, Union

import torch
from hypothesis.extra.numpy import arrays, array_shapes, from_dtype
from hypothesis import assume
import numpy as np


import hypothesis.strategies as st

NONZERO_CUT = 1e-6


def hilbert_space_dimensions(max_num_qubits, allow_0_qubits: bool = False):
    """Build a strategy that draws 2^k with k up to max_num_qubits."""
    def exponentiate(k):
        return 2 ** k

    if allow_0_qubits:
        min_num_qubits = 0
    else:
        min_num_qubits = 1
    return st.builds(exponentiate, st.integers(min_num_qubits, max_num_qubits))


torch_dtype_to_numpy = {
    torch.complex64: np.dtype('complex64'),
    torch.complex128: np.dtype('complex128'),
    torch.float32: np.dtype('float32'),
    torch.float64: np.dtype('float64'),
}

@st.composite
def tensors(
        draw,
        min_dims: int = 0,
        max_dims: int = 3,
        min_side: int = 1,
        max_side: int = 1024,
        dtype: np.dtype = np.dtype('complex64'),
        allow_infinity: bool = False
):
    """Strategy for torch.Tensor objects.

    Essentially wraps the standard array strategy with some assumptions."""
    shape_strategy = array_shapes(
        min_dims=min_dims,
        max_dims=max_dims,
        min_side=min_side,
        max_side=max_side
    )
    element_strategy = from_dtype(
        dtype=dtype,
        allow_nan=False,
        allow_infinity=allow_infinity
    )
    array = draw(arrays(
        dtype=dtype,
        shape=shape_strategy,
        elements=element_strategy
    ))

    return torch.tensor(array)


def real_and_complex_torch_dtypes(strictly_complex: bool):
    if strictly_complex:
        return st.just(torch.complex64)
    else:
        return st.sampled_from([torch.complex64, torch.float32])


def real_and_complex_np_dtypes(strictly_complex: bool):
    if strictly_complex:
        return st.just(np.dtype('complex64'))
    else:
        return st.sampled_from([np.dtype('complex64'), np.dtype('float32')])


@st.composite
def tensors_size_fixed(
        draw,
        shape: Union[Tuple[int], torch.Size],
        dtype: np.dtype = np.dtype('complex64'),
        allow_infinity: bool = False
):
    """Strategy for Tensor objects with definite shape and dtype."""
    shape = tuple(shape)
    element_strategy = from_dtype(
        dtype=dtype, allow_nan=False, allow_infinity=allow_infinity)

    array = draw(arrays(
        dtype=dtype,
        shape=shape,
        elements=element_strategy
    ))
    return torch.tensor(array)


def state_vector_size(
        batch_rank_limit: int = 3,
        batch_size_limit: int = 3,
        min_num_qubits: int = 1,
        max_num_qubits: int = 8,
):
    def construct_state_size(
            batch_dims: Tuple[int, ...],
            num_qubits: int
    ):
        return batch_dims + (2 ** num_qubits,)

    batch_dim_strategy = array_shapes(
        min_dims=0,
        max_dims=batch_rank_limit,
        min_side=1,
        max_side=batch_size_limit
    )
    return st.builds(
        construct_state_size,
        batch_dims=batch_dim_strategy,
        num_qubits=st.integers(min_num_qubits, max_num_qubits)
    )


@st.composite
def state_vectors(
        draw,
        batch_rank_limit: int = 3,
        batch_size_limit: int = 3,
        min_num_qubits: int = 1,
        max_num_qubits: int = 8,
        strictly_complex: bool = True
):
    """Strategy for drawing normalized states in vector layout.

    Args:
        draw: Hypothesis draw parameter--ignore this for testing.
        batch_rank_limit: The largest allowed number of batch indices.
        batch_size_limit: The largest allowed dimension for each batch index.
        min_num_qubits: The smallest number of qubits allowed.
        max_num_qubits: The largest number of qubits allowed.
        strictly_complex: When True, only draw complex dtypes.
    """
    dtype = draw(real_and_complex_np_dtypes(strictly_complex))
    size = draw(
        state_vector_size(
            batch_rank_limit=batch_rank_limit,
            batch_size_limit=batch_size_limit,
            min_num_qubits=min_num_qubits,
            max_num_qubits=max_num_qubits,
        ))
    element_strategy = from_dtype(
        dtype=dtype,
        min_value=-1, max_value=1,
        allow_nan=False, allow_infinity=False
    )
    array = draw(arrays(
        dtype=dtype,
        shape=size,
        elements=element_strategy
    ))

    # min_value and max_value don't avoid large arrays for complex cases.
    # TODO: This might have something to do with tests running very slowly.
    #   Need to further investigate!
    if np.iscomplexobj(array):
        assume((np.abs(array) < 2.).all())
    state = torch.tensor(array)

    state_norm = torch.norm(state, dim=-1)
    # Manually try to avoid zero states with a random shift
    if (state_norm.abs() < NONZERO_CUT).any():
        state = state + .1 * torch.rand_like(state)
        state_norm = torch.norm(state, dim=-1)

    # At this point, we "assume" that the states are all nonzero
    assume(
        (state_norm.abs() > NONZERO_CUT).all()
    )
    state_norm.unsqueeze_(-1)

    return state / state_norm


@st.composite
def state_vectors_with_metadata(
        draw,
        batch_rank_limit: int = 3,
        batch_size_limit: int = 3,
        min_num_qubits: int = 1,
        max_num_qubits: int = 8,
        strictly_complex: bool = True,
):
    num_qubits = draw(st.integers(min_num_qubits, max_num_qubits))
    state_vector = draw(state_vectors(
        batch_rank_limit,
        batch_size_limit,
        min_num_qubits=num_qubits,
        max_num_qubits=num_qubits,
        strictly_complex=strictly_complex
    ))
    batch_dims = state_vector.size()[:-1]

    return {
        'state_vector': state_vector,
        'num_qubits': num_qubits,
        'batch_dims': batch_dims,
        'dtype': state_vector.dtype,
    }


def state_vectors_no_batch(
        min_num_qubits: int = 1,
        max_num_qubits: int = 8,
        strictly_complex: bool = True
):
    """Convenience wrapper of state_vectors strategy with batching off.

    See documentation for state_vectors.
    """
    return state_vectors(
        batch_rank_limit=0,
        batch_size_limit=1,
        min_num_qubits=min_num_qubits,
        max_num_qubits=max_num_qubits,
        strictly_complex=strictly_complex
    )


@st.composite
def sizes(
        draw,
        min_num_dims: int = 0,
        max_num_dims: int = 3,
        min_index_range: int = 1,
        max_index_range: int = 6,
):
    """Strategy for torch.Size objects."""
    size = draw(st.lists(
        elements=st.integers(min_index_range, max_index_range),
        min_size=min_num_dims,
        max_size=max_num_dims
    ))
    return torch.Size(size)


@st.composite
def operators(
        draw,
        min_num_qubits: int = 1,
        max_num_qubits: int = 1,
        batch_max_num_indices: int = 3,
        batch_max_index_range: int = 5,
        strictly_complex: Optional[bool] = True,
        forced_dtype: Optional[torch.dtype] = None,
        nonzero: bool = False,
        max_abs: Optional[float] = None
):
    if forced_dtype is not None:
        dtype = torch_dtype_to_numpy[forced_dtype]
    else:
        dtype = draw(real_and_complex_np_dtypes(strictly_complex))
    batch_dims = draw(
        sizes(
            min_num_dims=0,
            max_num_dims=batch_max_num_indices,
            min_index_range=1,
            max_index_range=batch_max_index_range
        )
    )

    num_qubits = draw(st.integers(min_num_qubits, max_num_qubits))
    matrix_dims = (2 ** num_qubits, 2 ** num_qubits)

    all_dims = batch_dims + matrix_dims

    result = draw(tensors_size_fixed(shape=all_dims, dtype=dtype))
    if nonzero:
        assume(
            (result.abs() > NONZERO_CUT).all()
        )
    if max_abs is not None:
        assume(
            (result.abs() < max_abs).all()
        )

    return result


@st.composite
def operators_batch_fixed(
        draw,
        batch_dims: torch.Size = torch.Size([]),
        min_num_qubits: int = 1,
        max_num_qubits: int = 1,
        strictly_complex: bool = True,
        nonzero: bool = False,
        max_abs: Optional[float] = None
):
    dtype = draw(real_and_complex_np_dtypes(strictly_complex))

    num_qubits = draw(st.integers(min_num_qubits, max_num_qubits))
    matrix_dims = (2 ** num_qubits, 2 ** num_qubits)

    all_dims = batch_dims + matrix_dims

    result = draw(tensors_size_fixed(shape=all_dims, dtype=dtype))
    if nonzero:
        assume(
            (result.abs() > NONZERO_CUT).all()
        )

    if max_abs is not None:
        assume(
            (result.abs() < max_abs).all()
        )
    return result



import enum
import math
from typing import Union, List, Tuple

import torch


class StateLayout(str, enum.Enum):
    VECTOR = 'vector'
    TENSOR = 'tensor'


def count_qubits(state: torch.Tensor):
    """Get the number of qubits of a state in vector layout.

    This function works for complex and real cases and it is compatible
    with arbitrary batch dimensions.

    Args:
        state (Tensor): State (or batch of states) in vector layout.

    """
    length = state.size()[-1]
    num_bits = round(math.log2(length))
    if 2 ** num_bits != length:
        raise StateShapeError(data=state, expected_layout=StateLayout.VECTOR)
    return num_bits


# TODO: sharpen nomenclature around "gate", "operator", "matrix", etc.
def count_qubits_gate_matrix(gate: torch.Tensor):
    """Get the number of qubits that a gate matrix acts on.

    By convention, a gate matrix has the shape

        (*optional_batch_dims, 2^k, 2^k)

    where k is the number of qubits that the gate acts on. Note that
    k might be smaller than the number of qubits in a state that we
    are going to apply the gate to.
    """
    length = gate.size()[-1]
    num_bits = round(math.log2(length))
    if 2 ** num_bits != length:
        raise RuntimeError(f'Given gate matrix has size {gate.size()} which '
                           f'is not consistent with any number of qubits.')
    return num_bits


def hilbert_space_dim(state: torch.Tensor):
    """Get the dimension of the Hilbert space for a state in vector layout."""
    return 2 ** count_qubits(state)


def count_qubits_tensor(
        state_tensor: torch.Tensor,
        num_batch_dims: int,
):
    """Get the number of qubits of a state in tensor layout."""
    return state_tensor.dim() - num_batch_dims


def count_batch_dims_tensor(
        state_tensor: torch.Tensor,
        num_qubits: int,
):
    """Count the number of batch dimensions for a state in tensor layout.
    """
    return state_tensor.dim() - num_qubits


def get_qubit_indices(
        index: Union[int, List[int], Tuple[int, ...], torch.Tensor],
        state_tensor: torch.Tensor,
        num_qubits: int
):
    """Convert qubit indices 0, ..., n-1 to correct PyTorch tensor indices.

    Consider a state with two qubits in tensor layout and with one batch
    dimension. The size is (7, 2, 2) if the batch length is 7.
    The first dimension (with torch index 0) refers to the batch
    while the last two refer to qubits. If we assign the first and second
    qubits "qubit indices" 0 and 1 respectively, then the torch indices of
    (0 and 1) are (1 and 2) respectively.

    Examples:
        >>> state = torch.rand(2, 2)
        >>> get_qubit_indices(0, state, num_qubits=2)
        0
        >>> state = torch.rand(500, 17, 2, 2, 2)
        >>> get_qubit_indices(0, state, num_qubits=3)
        2
        >>> state = torch.rand(500, 17, 2, 2, 2)
        >>> get_qubit_indices([1, 0], state, num_qubits=3)
        [3, 2]
        >>> # Negative indices are not converted to positive:
        >>> state = torch.rand(500, 17, 2, 2, 2)
        >>> get_qubit_indices([-1, 0], state, num_qubits=3)
        [-1, 2]
    """
    batch_dims = state_tensor.dim() - num_qubits
    range_message = 'Expected index in {-num_qubits, ..., num_qubits - 1}.\n'
    range_message += f'Num_qubits: {num_qubits}, index: {index}.'

    convert = False
    if not isinstance(index, torch.Tensor):
        convert = True
        index = torch.tensor(index)

    if (index >= num_qubits).any() or (index < -num_qubits).any():
        raise ValueError(range_message)
    index = index + (index >= 0) * batch_dims
    if convert:
        return index.tolist()
    else:
        return index


# TODO: This function doesn't necessarily apply to states only so it would
#   be sensible to move it to utils or something.
def subset_roll_to_back(tensor: torch.Tensor, subset_num_dims):
    """Transpose front indices to the end of `tensor`.

    Examples:
        >>> x = torch.rand(4, 5, 6, 7, 8, 9)
        >>> subset_roll_to_back(x, 2).size()
        torch.Size([6, 7, 8, 9, 4, 5])
    """
    complement_range = range(subset_num_dims, tensor.dim())
    subset_range = range(subset_num_dims)
    perm = list(complement_range) + list(subset_range)
    return torch.permute(tensor, perm)


def subset_roll_to_front(tensor: torch.Tensor, subset_num_dims):
    """Transpose back indices to the front of `tensor`.

    Examples:
        >>> x = torch.rand(4, 5, 6, 7, 8, 9)
        >>> subset_roll_to_front(x, 2).size()
        torch.Size([8, 9, 4, 5, 6, 7])
    """
    subset_range = range(tensor.dim() - subset_num_dims, tensor.dim())
    complement_range = range(tensor.dim() - subset_num_dims)
    perm = list(subset_range) + list(complement_range)
    return torch.permute(tensor, perm)


class StateShapeError(ValueError):
    def __init__(
            self,
            data: torch.Tensor = None,
            expected_layout: StateLayout = None
    ):
        if data is not None:
            self.data_size = tuple(data.size())
        else:
            self.data_size = None
        if expected_layout is not None:
            expected_layout = StateLayout(expected_layout.lower())

        self.layout = expected_layout

    def __str__(self):
        message = (
            "There is a problem with the shape of a state.\n"
        )
        if self.layout is StateLayout.VECTOR:
            message += (
                "Expected VECTOR layout. This means that states are specified"
                " by a tensor with size (*optional_batch_dims, 2^num_qubits)\n"
        )
        if self.layout is StateLayout.TENSOR:
            message += (
                "Expected TENSOR layout. This means that states are specified"
                " by a tensor with size\n"
                "    (*optional_batch_dims, 2, 2, ..., 2)\n"
                "where the number of 2's is the number of qubits.\n"
            )
        if self.data_size is not None:
            message += f"Given state has size {self.data_size}"
        return message



import unittest
import torch
import unitair.simulation as sim
import unitair.gates as gates
import unitair.initializations as initializations
import unitair.states as states


class UnitaryOperatorCases(unittest.TestCase):
    def test_phases(self):

        def basic_phase(angles_: torch.Tensor, state_: torch.Tensor):
            """Multiply the jth dimension of state by e^(-i angles_j)."""
            cos = angles_.cos()
            sin = angles_.sin()
            return torch.complex(
                cos * state.real + sin * state.imag,
                -sin * state.real + cos * state.imag
            )

        state = initializations.rand_state(5)
        angles = torch.rand(2 ** 5)

        rotated = sim.apply_phase(angles, state)
        rotated_basic = basic_phase(angles, state)

        self.assertTrue((torch.isclose(rotated, rotated_basic)).all())

        state_batch = torch.rand(4, 2, 2 ** 3)
        angles = torch.rand(4, 2, 2 ** 3)
        rotated = sim.apply_phase(angles, state_batch)

        single_entry_rotated = sim.apply_phase(angles[2, 0], state_batch[2, 0])

        self.assertEqual(rotated.size(), torch.Size([4, 2, 2 ** 3]))
        self.assertTrue((rotated[2, 0] == single_entry_rotated).all())

    def test_apply_to_apply_all(self):
        for n in (1, 7):

            rot = gates.exp_x(.3)

            init_state = initializations.rand_state(n)
            final_1 = sim.apply_all_qubits(rot, init_state)

            final_2 = sim.apply_to_qubits(
                    operators=[rot for _ in range(n)],
                    qubits=range(n),
                    state=init_state
                )

            self.assertTrue((final_1 == final_2).all())

    def test_apply_last(self):
        for n in range(1, 6):

            op = torch.rand(2, 2, dtype=torch.complex64)

            state = initializations.rand_state(n)
            final_1 = sim.act_last_qubit(op, state)
            final_2 = sim.apply_to_qubits([op], [n - 1], state)

            self.assertTrue((final_1 == final_2).all())

        for n in range(1, 5):
            op = torch.rand(2, 2, dtype=torch.complex64)

            state = initializations.rand_state(n)
            final_1 = sim.act_last_qubit(op, state)
            final_2 = sim.apply_to_qubits([op], [n - 1], state)

            self.assertTrue((final_1 == final_2).all())

        op = torch.rand(2, 2, dtype=torch.complex64)
        state_batch = torch.rand(4, 3, 2, 2, 2, 2, 2, dtype=torch.complex64)
        final_batch = sim.act_last_qubit(op, state_batch)
        same = torch.allclose(
            final_batch[2, 2],
            sim.act_last_qubit(op, state_batch[2, 2])
        )
        self.assertTrue(
            same,
            msg='Problem with batching behavior for act_last_qubit.'
        )
        same = torch.allclose(
            final_batch[0, 1],
            sim.act_last_qubit(op, state_batch[0, 1])
        )
        self.assertTrue(
            same,
            msg='Problem with batching behavior for act_last_qubit.'
        )

    def test_swap(self):
        # First check some invariant examples.
        state = torch.tensor([.3, 0., 0., -0.13])
        swapped = sim.swap(state, qubit_pair=(0, 1))
        self.assertTrue(swapped.allclose(state))
        # complex invariant example
        state = torch.tensor([2.24 + .3j, 1. + .2j, 1. + .2j, 73. - .13j])
        swapped = sim.swap(state, qubit_pair=(0, 1))
        self.assertTrue(swapped.allclose(state))

        # random example with five qubits:
        state = initializations.rand_state(5)
        swapped = sim.swap(state, (0, 4))
        state_tensor = states.to_tensor_layout(state)
        manual_swap = state_tensor.transpose(0, 4)
        manual_swap = states.to_vector_layout(manual_swap, num_qubits=5)
        self.assertTrue(swapped.allclose(manual_swap))

        # Now a complicated example with batch dimensions for four qubits:
        state = torch.rand(3, 1, 3, 2 ** 4, dtype=torch.complex64)
        swapped = sim.swap(state, (1, 2))

        state_tensor = states.to_tensor_layout(state)
        manual_swap = state_tensor.transpose(4, 5)
        manual_swap = states.to_vector_layout(manual_swap, num_qubits=4)
        self.assertTrue(swapped.allclose(manual_swap))

        # Check batching behavior
        state = torch.rand(2, 3, 2 ** 4, dtype=torch.complex64)
        swapped = sim.swap(state, qubit_pair=(3, 1))
        swapped_entry = sim.swap(state[0, 1], qubit_pair=(3, 1))
        self.assertTrue(swapped[0, 1].allclose(swapped_entry))
        swapped_entry = sim.swap(state[1, 1], qubit_pair=(3, 1))
        self.assertTrue(swapped[1, 1].allclose(swapped_entry))



if __name__ == '__main__':
    unittest.main()



from .operations import apply_phase
from .operations import apply_operator
from .operations import apply_to_qubits
from .operations import apply_all_qubits
from .operations import act_last_qubit
from .operations import swap
from .operations import multi_cz
from .operations import multi_controlled_x, multi_controlled_z
from .operations import act_first_qubits
from .measurement import measure




from . import states
from . import simulation

from .states.shapes import StateLayout
from .initializations import unit_vector, uniform_superposition, rand_state

from .states import count_qubits, hilbert_space_dim
from .states import diag_expectation_value
from .states import inner_product, norm_squared, abs_squared

VECTOR_LAYOUT = states.shapes.StateLayout.VECTOR
TENSOR_LAYOUT = states.shapes.StateLayout.TENSOR



import torch
from typing import Union, Optional
from .gate_construction import parameterized_gate, constant_gate


@parameterized_gate(strictly_complex=True)
def exp_x(
        angle: Union[torch.Tensor, float],
        dtype: torch.dtype = torch.complex64
):
    """Get the operator e^(-i angle X).

    PyTorch device is inherited from the device of `angle`. If `angle` is
    a float, CPU is used.

    Args:
        angle: Tensor with size (batch_length,) or just ().

        dtype: Data type for the gate. Required to be complex.

    Returns: Tensor with size (batch_length, 2, 2, 2) or (2, 2, 2) if there
        is no batch dimension. The (2, 2, 2) is such that the first dimension
        means the real and imaginary parts and the last two dimension are
        the matrices of the real an imaginary parts of the gates.
    """
    angle = angle.to(dtype=dtype)
    cos = torch.cos(angle)
    i_sin = torch.sin(angle) * 1.j
    return [
        [[cos, -i_sin],
         [-i_sin, cos]],
    ]


@parameterized_gate(strictly_complex=True)
def exp_y(
        angle: Union[torch.Tensor, float],
        dtype: torch.dtype = torch.complex64
):
    """Get the operator e^(-i angle Y).

    PyTorch device is inherited from the device of `angle`. If `angle` is
    a float, CPU is used.

    Args:
        angle: Tensor with size (batch_length,) or just ().

        dtype: Data type for the gate. Required to be complex.

    Returns: Tensor with size (batch_length, 2, 2, 2) or (2, 2, 2) if there
        is no batch dimension. The (2, 2, 2) is such that the first dimension
        means the real and imaginary parts and the last two dimension are
        the matrices of the real an imaginary parts of the gates.
    """
    angle = angle.to(dtype=dtype)
    cos = torch.cos(angle)
    sin = torch.sin(angle)
    return [
        [[cos, -sin],
         [sin, cos]],
    ]


@parameterized_gate(strictly_complex=True)
def exp_z(
        angle: Union[torch.Tensor, float],
        dtype: torch.dtype = torch.complex64
):
    """Get the operator e^(-i angle Z).

    PyTorch device is inherited from the device of `angle`. If `angle` is
    a float, CPU is used.

    Args:
        angle: Tensor with size (batch_length,) or just ().

        dtype: Data type for the gate. Required to be complex.

    Returns: Tensor with size (batch_length, 2, 2, 2) or (2, 2, 2) if there
        is no batch dimension. The (2, 2, 2) is such that the first dimension
        means the real and imaginary parts and the last two dimension are
        the matrices of the real an imaginary parts of the gates.
    """
    angle = angle.to(dtype=dtype)
    cos = torch.cos(angle)
    i_sin = torch.sin(angle) * 1.j
    zero = torch.zeros(angle.size(), device=angle.device)
    return [
        [[cos - i_sin, zero],
         [zero, cos + i_sin]],
    ]


@constant_gate(strictly_complex=True)
def hadamard(
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None
):
    """Get the Hadamard gate.

    Args:
        device: If the torch device is not specified, CPU is used.

        dtype: torch.dtype for the result.
            When not specified, torch.complex64 is used.
    """
    val = 2. ** (-.5) + 0.j
    return [
        [val, val],
        [val, -val]
    ]


@constant_gate(strictly_complex=True)
def pauli_x(
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None
):
    """Get the Pauli X gate.

    Args:
        device: If the torch device is not specified, CPU is used.

        dtype: torch.dtype for the result.
            When not specified, torch.complex64 is used.
    """
    return [
        [0. + 0.j, 1. + 0.j],
        [1. + 0.j, 0. + 0.j]
    ]


@constant_gate(strictly_complex=True)
def pauli_y(
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None
):
    """Get the Pauli Y gate.

    Args:
        device: If the torch device is not specified, CPU is used.
        dtype: torch.dtype for the result.
            When not specified, torch.complex64 is used.
    """
    return [
        [0., -1.j],
        [1.j, 0.]
    ]


@constant_gate(strictly_complex=True)
def pauli_z(
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None
):
    """Get the Pauli Z gate.

    Args:
        device: If the torch device is not specified, CPU is used.

        dtype: torch.dtype for the result.
            When not specified, torch.complex64 is used.
    """
    return [
        [1. + 0.j, 0.j],
        [0.j, -1. + 0.j]
    ]


@constant_gate(strictly_complex=True)
def cnot(
        device: Optional[torch.device] = None,
        dtype: Optional[torch.device] = None
):
    """Get the CNOT gate.

    Args:
        device: If the torch device is not specified, CPU is used.

        dtype: torch.dtype for the result.
            When not specified, torch.complex64 is used.
    """
    return [
        [1. + 0.j, 0. + 0.j, 0. + 0.j, 0. + 0.j],
        [0. + 0.j, 1. + 0.j, 0. + 0.j, 0. + 0.j],
        [0. + 0.j, 0. + 0.j, 0. + 0.j, 1. + 0.j],
        [0. + 0.j, 0. + 0.j, 1. + 0.j, 0. + 0.j]
    ]


# `cx` is an alias for `cnot`.
cx = cnot


@constant_gate(strictly_complex=True)
def cz(
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None
):
    """Get the controlled-Z gate.

    Args:
        device: If the torch device is not specified, CPU is used.

        dtype: torch.dtype for the result.
            When not specified, torch.complex64 is used.
    """
    return [
        [1. + 0.j, 0. + 0.j, 0. + 0.j, 0. + 0.j],
        [0. + 0.j, 1. + 0.j, 0. + 0.j, 0. + 0.j],
        [0. + 0.j, 0. + 0.j, 1. + 0.j, 0. + 0.j],
        [0. + 0.j, 0. + 0.j, 0. + 0.j, -1. + 0.j]
    ]

